import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQml.Models 2.2
import QtQuick.Window 2.2
import QtQuick.Layouts 1.3
import QtQml 2.12
import org.ukui.tablet 1.0
import org.ukui.qqc2style.private 1.0 as StylePrivate
import "launcherItemViewController.js" as Controller
import "createUninstallDialog.js" as UninstallJs
import "."

Item {
    id: root
    function listViewItemCount() {
        return listView.count
    }

    property var viewMoel: null
    property bool dragEnable: true
    property int groupId: -1
    property real cellSize: Style.widthScaleFactor * 120
    property real cellContentMargin: Style.widthScaleFactor * 65
    property var pageContainer: listView
    property bool flicking: false

    property real cellWidth: root.width/4
    property real cellHeight: root.height/4
    property real iconSize: 96 * Style.tabletScaleFactor

    property string iconPromptBox
    property string textPromptBox
    property string tarDesktopName
    property bool allowGroupItemToTaskbar: false

    signal startDrag(bool start)
    signal containerClicked()
    // 点击分组内的“添加”按钮显示添加应用的界面
    signal showAddAppWindow(int groupId, int pageIndex)
    signal focusStateChanged()
    signal placeholderAdded()
    signal changeGroupHeight()
    signal launcherItemToDesktop()
    property bool groupAnimationState: true

    clip: true
    ListView {
        id: listView
        anchors.fill: parent
        spacing: 100
        snapMode: ListView.SnapOneItem
        model: viewMoel
        delegate: listViewItemDelegate
        orientation: ListView.Horizontal
        highlightRangeMode: ListView.StrictlyEnforceRange
        boundsBehavior:Flickable.StopAtBounds
        highlightMoveDuration: 500
        highlightFollowsCurrentItem: true
        cacheBuffer: parent.width * listView.count
        clip: false
        focus: true
    }
    Component {
        id: listViewItemDelegate
        GridView {
            id: gridView
            width: listView.width
            height: listView.height
            cellWidth: root.cellWidth
            cellHeight: root.cellHeight
            model: modelData
            delegate: gridviewItemDelegate
            addDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
            move: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
            moveDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
            remove: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
            removeDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 100 } }
            displaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
            interactive: false
            MouseArea {
                anchors.fill: parent
                z: -1
                pressAndHoldInterval: Style.pressIntervalToEditable
                onClicked: {
                    // do to edit focus
                    root.focusStateChanged()
                    modelManager.setAllItemsState(LauncherItem.Normal)
                    changeGroupHeight()
                }
                onPressAndHold: {
                    startDrag(true)
                    modelManager.setAllItemsState(LauncherItem.Editable)
                    changeGroupHeight()
                }
            }
        }
    }

    Component {
        id: gridviewItemDelegate
        DropArea {
            width: modelData.columnSpan * root.cellWidth
            height: modelData.rowSpan * root.cellHeight
            opacity: 1
            property int visualIndex: index
            property int pageIndex: modelData.pageIndex
            property int rowSpan: modelData.columnSpan
            property int columnSpan: modelData.columnSpan
            property int type: modelData.type
            Binding { target: itemRoot; property: "visualIndex"; value: visualIndex }
            Binding { target: itemRoot; property: "pageIndex"; value: pageIndex }
            Binding { target: itemRoot; property: "rowSpan"; value: rowSpan }
            Binding { target: itemRoot; property: "columnSpan"; value: columnSpan }

            onEntered: {
                console.log("LauncherItemView onEntered",
                            drag.source.pageIndex, pageIndex, modelData.pageIndex, drag.source.visualIndex, visualIndex, dragEnable)
                if (!dragEnable || modelData.type === LauncherItem.AddIcon /*|| drag.source.groupId !== root.groupId*/) {
                    return
                }
                if (drag.source.visualIndex === undefined) {
                    return
                }
                if (drag.source.pageIndex !== pageIndex || drag.source.groupId !== root.groupId) {
                    if (!groupAnimationState) {
                        modelManager.addPlaceholderItem(
                                    visualIndex, pageIndex,
                                    drag.source.rowSpan,
                                    drag.source.columnSpan,
                                    groupId)
                    }
                    placeholderAdded()
                } else {
                    modelManager.moveItem(drag.source.visualIndex, visualIndex, pageIndex, groupId)
                    modelManager.removePlaceholderItem();
                }
            }
            onExited: {
                modelManager.setItemState(LauncherItem.Editable, modelData.id)
            }

            Item {
                id: itemRoot
                width: iconSize//modelData.columnSpan * root.cellWidth
                height: iconSize//modelData.rowSpan * root.cellHeight
                anchors.horizontalCenter: parent.horizontalCenter
                anchors.top: parent.top
                anchors.topMargin: (modelData.columnSpan * root.cellWidth - iconSize)/2 * Style.tabletScaleFactor
                clip: false

                property int visualIndex: 0
                property int pageIndex: 0
                property int rowSpan: 1
                property int columnSpan: 1
                property int groupId: root.groupId
                property int itemId: modelData.id
                property bool mouseReleased: false
                property int type: modelData.type
                property var dragAppIcon: item
                property var dragAppName: itemName

                Drag.active: dragMousearea.drag.active
                Drag.source: itemRoot
                Drag.hotSpot.x: itemRoot.width / 2
                Drag.hotSpot.y: itemRoot.height / 2

                states: [
                    State {
                        when: itemRoot.Drag.active
                        ParentChange {
                            target: itemRoot
                            parent: topLayer
                        }

                        AnchorChanges {
                            target: itemRoot
                            anchors.horizontalCenter: undefined
                            anchors.top: undefined
                        }

                        PropertyChanges {
                            target: item
                            iconState: LauncherItem.Dragging
                            iconWidth: 96 * Style.tabletScaleFactor * 1.1
                        }

                        PropertyChanges {
                            target: itemRoot
                            width: 96 * Style.tabletScaleFactor * 1.1
                            height: 96 * Style.tabletScaleFactor * 1.1
                        }
                    }
                ]

                MouseArea {
                    id: dragMousearea
                    anchors.fill: parent
                    hoverEnabled: false
                    pressAndHoldInterval: modelData.state === LauncherItem.Editable ?
                                              100 : Style.pressIntervalToEditable
                    property bool isScrolling: false
                    property bool isChangeListviewPage: false
                    property bool groupHeightChanged: false
                    property point startDragPoint: Qt.point(0, 0)

                    Timer {
                        id: changeListviewPage
                        interval: 300
                        onTriggered: {
                            dragMousearea.isChangeListviewPage = true
                        }
                    }
                    Timer {
                        id: itemToDesktopTimer
                        interval: 600
                        onTriggered: {
                            GroupController.minimizeGroup()
                            launcherItemToDesktop()
                            itemToDesktopTimer.stop()
                        }
                    }

                    onPressAndHold: {
                        if (modelData.type === LauncherItem.AddIcon) {
                            return;
                        }
                        startDrag(true)
                        modelManager.setAllItemsState(LauncherItem.Editable)
                        dragMousearea.Drag.active = true
                        dragMousearea.drag.target = itemRoot
                        startDragPoint = mapToGlobal(itemRoot.x, itemRoot.y)
                        console.log("startDragPoint", startDragPoint)
                    }

                    onClicked: {
                        if (modelData.type === LauncherItem.AddIcon) {
                            showAddAppWindow(root.groupId, modelData.pageIndex)
                        } else if (modelData.type === LauncherItem.AppIcon && modelData.state === LauncherItem.Normal) {
                            appManager.runApplication(modelData.desktopFile)
                        }
                    }
                    onPositionChanged:  {
                        Controller.appItemPositionChanged(dragMousearea,
                                                          changeListviewPage,
                                                          itemToDesktopTimer,
                                                          item,
                                                          itemRoot,
                                                          GroupController.groupContainer,
                                                          GroupController.groupContainerId)
                    }
                    onPressed: {
                        if (modelData.type === LauncherItem.AppIcon) {
                            itemRoot.mouseReleased = false
                        }
                        if (modelData.type === LauncherItem.AddIcon) {
                            iconpressAnimation.target = item
                            iconpressAnimation.start()
                        }
                    }

                    onReleased: {
                        console.log("Launcher item view release.")
                        itemToDesktopTimer.stop()
                        itemRoot.mouseReleased = true
                        if (modelData.type === LauncherItem.AddIcon) {
                            iconreleseAnimation.target = item
                            iconreleseAnimation.start()
                            return
                        }
                        startDrag(false)
                        isScrolling = false
                        dragMousearea.Drag.active = false
                        dragMousearea.drag.target = null
                        changeGroupHeight()
                        groupHeightChanged = false
                        if (GroupController.aboutToInsertGroup && GroupController.aboveItem.groupId !== GroupController.popupedGroupId) {
                            GroupController.insertItemToGroup(GroupController.aboveItem.visualIndex, GroupController.aboveItem.pageIndex,
                                                              GroupController.aboveItem.itemId, GroupController.aboveItem.dragAppIcon,
                                                              GroupController.aboveItem.dragAppName)
                            modelManager.updateCachedData(GroupController.popupedGroupId)
                        } else if (taskbar.replayAllowed) {
                            if (modelData.type === LauncherItem.AppIcon) {
                                if (modelManager.replacePlaceholderItemTaskbar(visualIndex, pageIndex, groupId, 0)) {
                                    modelManager.updateCachedData(groupId)
                                }
                            }
                            taskbar.replayAllowed = false
                        } else {
                            modelManager.replacePlaceholderItem(itemRoot.visualIndex, itemRoot.pageIndex, groupId)
                        }
                        modelManager.updateCachedData()
                        modelManager.updateCachedData(0)
                    }
                }

                AppIcon {
                    id: item
                    iconWidth: iconSize
                    visible: modelData.type !== LauncherItem.Placeholder
                    disableDelete: !modelManager.groupEditable(root.groupId)
                    opacity: {
                        if (modelData.type === LauncherItem.AppIcon) {
                            if (listView.moving || dragMousearea.Drag.active) {
                                return 1
                            }
                            if (dragMousearea.pressed && !dragMousearea.Drag.active) {
                                return 0.5
                            }
                        }
                        return 1
                    }
                    icon: {
                        switch (modelData.type) {
                        case LauncherItem.AppIcon: return "image://icons/" + modelData.icon
                        case LauncherItem.AddIcon:
                            if (dragMousearea.pressed) {
                                return "qrc:/ui/images/list-add-click.svg";
                            } else {
                                return "qrc:/ui/images/list-add-default.svg";
                            }
                        default: return ""
                        }

                    }
                    iconState: modelData.state
                    anchors {
                        horizontalCenter: parent.horizontalCenter;
                        verticalCenter: parent.verticalCenter
                    }
                    onDeleteIcon: {
                        appIcon.visible = true
                        UninstallJs.createUnInstallDialog(root, modelData.desktopFile, modelData.name, modelData.icon)
                        textPromptBox = model.modelData.name
                        iconPromptBox = modelData.icon
                    }
                    onIconStateChanged: {
                        if (modelData.type === LauncherItem.AddIcon) {
                            if (iconState === LauncherItem.Normal) {
                                item.visible = true
                                itemName.visible = true
                                dragMousearea.enabled = true
                            } else {
                                item.visible = false
                                itemName.visible = false
                                dragMousearea.enabled = false
                            }
                        }
                    }
                }
                Text {
                    id: itemName
                    width: item.width + 40 * Style.widthScaleFactor
                    anchors { top: item.bottom; topMargin: 8 * Style.tabletScaleFactor
                        horizontalCenter: item.horizontalCenter }
                    horizontalAlignment: Text.AlignHCenter
                    color: "black"
                    text: model.modelData.name
                    elide: Text.ElideRight
                    font: StylePrivate.StyleHelper.font
                    smooth: true
                    visible: modelData.type !== LauncherItem.Placeholder &&
                             modelData.nameVisible &&
                             item.iconState !== LauncherItem.Dragging
                }
                Component.onCompleted: {
                    if (viewMoel === null) {
                        console.warn("AppContainer: The model is null, please assign a model!")
                    }
                }
            }
        }
    }
    ScaleAnimator {
        id: iconpressAnimation
        from: 1
        to: 0.8
        duration: 150
    }
    ScaleAnimator {
        id: iconreleseAnimation
        from: 0.8
        to: 1
        duration: 150
    }

    onContainerClicked: {
        modelManager.setItemState(LauncherItem.Normal, groupId)
    }
}






