import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Window 2.0
import QtQml.Models 2.2
import QtQuick.Window 2.2
import QtQuick.Layouts 1.3
import QtQml 2.12
import org.ukui.tablet 1.0
import org.ukui.qqc2style.private 1.0 as StylePrivate
import "createUninstallDialog.js" as UninstallJs
import "."

Item {
    id: root
    property var viewMoel: null
    property bool dragEnable: true
    property real gridCellWidth: Style.tabletScaleFactor * 254
    property real gridCellHeight: Style.tabletScaleFactor * 206
    property real cellContentMargin: Style.tabletScaleFactor * 20
    property Item pluginAnchorItem: undefined
    property var rootWindow: undefined
    property var pageContainer: listView
    property int groupId: -1
    property var appImage
    property var apptext
    property int appNameWidth: 128 * Style.tabletScaleFactor
    property alias taskBar: taskbar

    signal startDrag(bool started)
    signal showAddAppWindow(int groupId, int pageIndex)
    signal titleFocusChanged()

    function listViewItemCount() {
        return listView.count
    }

    clip: false

    ListView {
        id: listView
        signal movingChanged(bool m)
        anchors.fill: parent
        spacing: 40
        snapMode: ListView.SnapOneItem
        model: viewMoel
        delegate: listViewItemDelegate
        orientation: ListView.Horizontal
        highlightRangeMode: ListView.StrictlyEnforceRange
        highlightFollowsCurrentItem: true
        highlightMoveDuration: 100
        maximumFlickVelocity: 8000
        cacheBuffer: root.width * listView.count
        clip: false
        focus: true
        onHorizontalVelocityChanged: movingChanged(horizontalVelocity)
    }
    Component {
        id: listViewItemDelegate
        Item {
            width: listView.width
            height: listView.height
            KGridView {
                id: gridView
                width: Screen.width > Screen.height ?
                           240 * 6 * Style.tabletScaleFactor + 1 :
                           240 * 4 * Style.tabletScaleFactor + 1
                y: 75 * Style.tabletScaleFactor
                height: parent.height
                anchors.horizontalCenter: parent.horizontalCenter
                cellWidth: Style.appIconGridWidth//gridCellWidth
                cellHeight: Style.appIconGridHeight//gridCellHeight
                model: modelData
                interactive: false
                delegate: gridviewItemDelegate
                addDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
                move: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
                moveDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
                remove: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
                removeDisplaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }
                displaced: Transition { NumberAnimation { properties: "x,y"; duration: 200 } }

                MouseArea {
                    id: desktopMouseArea
                    property bool editableTrigged: false
                    property point startPoint: Qt.point(0, 0)
                    anchors.fill: parent
                    z: -1
                    pressAndHoldInterval: Style.pressIntervalToEditable
                    onPositionChanged: {
                        if (Math.abs(startPoint.x - mouseX) > 20 * Style.tabletScaleFactor ||
                                Math.abs(startPoint.y - mouseY) > 20 * Style.tabletScaleFactor) {
                            editableTrigged = false
                        }
                    }
                    onPressed: {
                        startPoint = Qt.point(mouseX, mouseY)
                        editableTrigged = true
                    }
                    onClicked: {
                        //do to edit focus
                        modelManager.setAllItemsState(LauncherItem.Normal)
                    }
                    onPressAndHold: {
                        if (!editableTrigged) {
                            return
                        }
                        startDrag(true)
                        modelManager.setAllItemsState(LauncherItem.Editable)
                    }
                }
            }
        }
    }

    Component {
        id: gridviewItemDelegate
        DropArea {
            id: delegateContent
            property int visualIndex: index
            property int pageIndex: modelData.pageIndex
            property int rowSpan: modelData.rowSpan
            property int columnSpan: modelData.columnSpan
            property int itemId: modelData.id
            property bool mouseAreaReleased: false
            property int type: modelData.type

            width: modelData.columnSpan *  Style.appIconGridWidth
            height: modelData.rowSpan *  Style.appIconGridHeight
            opacity: 1
            enabled: !GroupController.groupPopuped

            Binding { target: itemRoot; property: "visualIndex"; value: visualIndex }
            Binding { target: itemRoot; property: "pageIndex"; value: pageIndex }
            Binding { target: itemRoot; property: "rowSpan"; value: rowSpan }
            Binding { target: itemRoot; property: "columnSpan"; value: columnSpan }
            Binding { target: itemRoot; property: "itemId"; value: itemId }

            Component.onCompleted: {
                itemRoot.visualIndex = index
            }

            Timer {
                id: popUpTimer
                property int dragSourceId: 0
                property int timeOutCount: 0
                interval: 600
                repeat: true

                onTriggered: {
                    ++ timeOutCount
                    if (timeOutCount == 1) {
                        modelManager.setItemState(LauncherItem.Under, modelData.id)
                        GroupController.aboutToInsertItemToGroup(itemRoot, drag.source)
                    }
                    if (timeOutCount >= 2) {
                        if (drag.source !== null) {
                            GroupController.popupGroup(appIcon, true)
                            //删除桌面占位符
                            modelManager.removePlaceholderItem()
                            popUpTimer.stop()
                        }
                    }
                }
                onRunningChanged: {
                    if (!running) {
                        timeOutCount = 0
                    }
                }
            }

            onEntered: {
                if (!dragEnable ||
                        (modelData.type !== LauncherItem.AppIcon && modelData.type !== LauncherItem.Group) ||
                        drag.source.groupId === 0 || drag.source.visualIndex === visualIndex ||
                        drag.source.itemId === undefined || drag.source.type === LauncherItem.Group ||
                        drag.source.type === LauncherItem.Widget) {
                    return
                }
                popUpTimer.start()
            }
            onExited: {
                if (!GroupController.groupPopuped) {
                    // 鼠标release事件也会导致进入onExited
                    if (!mouseAreaReleased) {
                        GroupController.canceInsertToGroup()
                        modelManager.setItemState(LauncherItem.Editable, modelData.id)
                        popUpTimer.stop()
                    } else {
                        mouseAreaReleased = false
                    }
                }
            }
            DropArea {
                id: moveDropArea
                width: modelData.columnSpan * 80 + (modelData.columnSpan-1) * 160
                height: modelData.rowSpan * 80 + (modelData.rowSpan-1) * 114

                anchors.centerIn: parent
                onEntered: {
                    if (!dragEnable || modelData.type === LauncherItem.AddIcon /*||
                            drag.source.groupId !== root.groupId || drag.source.visualIndex === visualIndex*/) {
                        return
                    }
                    if (drag.source.visualIndex === undefined || drag.source.visualIndex === -1) {
                        return
                    }
                    popUpTimer.stop()
                    if (drag.source.pageIndex !== pageIndex || drag.source.groupId !== root.groupId) {
                        modelManager.addPlaceholderItem(visualIndex, pageIndex, drag.source.rowSpan, drag.source.columnSpan)
                    } else {
                        console.log("about to move item...", drag.source.visualIndex, visualIndex, pageIndex)
                        modelManager.moveItem(drag.source.visualIndex, visualIndex, pageIndex)
                        modelManager.removePlaceholderItem();
                    }
                }
                onExited: {
                    modelManager.setItemState(LauncherItem.Editable, modelData.id)
                }
            }

            Item {
                id: itemRoot
                width: modelData.columnSpan * Style.appIconGridWidth
                height: modelData.rowSpan * Style.appIconGridHeight
                anchors.horizontalCenter: parent.horizontalCenter
                anchors.top: parent.top
                clip: false

                // 方便在拖动时获取数据
                property int visualIndex: 0
                property int pageIndex: 0
                property int rowSpan: modelData.rowSpan
                property int columnSpan: modelData.columnSpan
                property int itemId: modelData.id
                property int groupId: root.groupId
                property int type: modelData.type

                Drag.active: dragMousearea.drag.active && dragEnable
                Drag.source: itemRoot
                Drag.hotSpot.x: Style.appIconGridWidth / 2
                Drag.hotSpot.y: Style.appIconGridHeight / 2


                states: [
                    State {
                        when: itemRoot.Drag.active
                        ParentChange {
                            target: itemRoot
                            parent: topLayer
                        }

                        AnchorChanges {
                            target: itemRoot
                            anchors.horizontalCenter: undefined
                            anchors.top: undefined
                        }
                        PropertyChanges {
                            target: appIcon
                            width: appIcon.width * 1.1
                            height: appIcon.height * 1.1
                            containerState: LauncherItem.Dragging
                        }
                    }
                ]

                MouseArea {
                    id: dragMousearea
                    anchors.centerIn: parent
                    width: modelData.columnSpan * Style.appIconGridWidth - Style.tabletScaleFactor * 40
                    height: modelData.rowSpan * Style.appIconGridHeight - Style.tabletScaleFactor * 40
                    hoverEnabled: true
                    Drag.active: false
                    drag.target: undefined
                    pressAndHoldInterval: modelData.state === LauncherItem.Editable ?
                                              100 : Style.pressIntervalToEditable
                    property bool isScrolling: false
                    property var mapContainer: GroupController.groupContainer
                    property bool isChangeContainerPage: false

                    Timer {
                        id: scrollingTimer
                        interval: 600
                        onTriggered: dragMousearea.isScrolling = false
                    }
                    Timer {
                        id: changeContainerPageTimer
                        interval: 300
                        onTriggered: {
                            dragMousearea.isChangeContainerPage = true
                        }
                    }
                    Timer {
                        id: itemToDesktopTimer
                        interval: 600
                        onTriggered: {
                            GroupController.minimizeGroup()
                        }
                    }
                    onPressed: {
                        if (dragMousearea.Drag.active) {
                            dragMousearea.drag.target = itemRoot
                        }
                    }
                    onPressAndHold: {
                        modelManager.setAllItemsState(LauncherItem.Editable)
                        dragMousearea.Drag.active = true
                        dragMousearea.drag.target = itemRoot
                        startDrag(true)
                    }
                    onClicked: {
                        switch (modelData.type) {
                        case LauncherItem.AppIcon: {
                            if (modelData.state === LauncherItem.Normal){
                                appManager.runApplication(modelData.desktopFile)
                            }
                            break;
                        }

                        case LauncherItem.Group: {
                            if (modelData.state !== LauncherItem.Popup) {
                                GroupController.popupGroup(appIcon)
                            }
                            break;
                        }
                        default: break;
                        }
                        root.titleFocusChanged()
                    }
                    onPositionChanged:  {
                        if (taskbar.replayAllowed) {
                            appIcon.width = 64 * Style.tabletScaleFactor
                            appIcon.height = 64 * Style.tabletScaleFactor
                        }
                        if (!taskbar.replayAllowed && itemRoot.Drag.active) {
                            appIcon.width = type === LauncherItem.Group ? (88*Style.tabletScaleFactor*1.1) : (96*Style.tabletScaleFactor*1.1)
                            appIcon.height = type === LauncherItem.Group ? (88*Style.tabletScaleFactor*1.1) : (96*Style.tabletScaleFactor*1.1)
                        }

                        var itemPos = item.mapToItem(listView, 0, 0)


                        //listView.width - ((item.width - moveDropArea.width) / 2 + moveDropArea.width) * Style.tabletScaleFactor
                        //item.width - (item.width - moveDropArea.width)/2
                        if ((itemPos.x % listView.width) >= (listView.width - (item.width / 2 + moveDropArea.width / 2) * Style.tabletScaleFactor) &&
                                dragMousearea.Drag.active &&!isScrolling && !GroupController.groupPopuped) {
                            if (listView.currentIndex < listView.count - 1) {
                                listView.currentIndex ++
                                isScrolling = true
                                scrollingTimer.start()
                                modelManager.removePlaceholderItem()
                                // 拖动至不同页时才添加 placeholder item
                                if (drag.target.pageIndex !== listView.currentIndex) {
                                    var curPageItemCount = modelManager.groupPageItemCount(-1, listView.currentIndex);
                                    var index = curPageItemCount > 0 ? curPageItemCount : 0
                                    modelManager.addPlaceholderItem(index, listView.currentIndex, drag.target.rowSpan,
                                                drag.target.columnSpan)
                                }
                            } else {
                                item.x = 0
                            }
                        } else if ((itemPos.x % listView.width < -item.width/2 + moveDropArea.width / 2 * Style.tabletScaleFactor) &&
                                   dragMousearea.Drag.active && !isScrolling && !GroupController.groupPopuped) {
                            if (listView.currentIndex > 0) {
                                listView.currentIndex --
                                isScrolling = true
                                scrollingTimer.start()
                                modelManager.removePlaceholderItem()
                                if (drag.target.pageIndex !== listView.currentIndex) {
                                    curPageItemCount = modelManager.groupPageItemCount(-1, listView.currentIndex);
                                    index = curPageItemCount > 0 ? curPageItemCount : 0
                                    modelManager.addPlaceholderItem(index, listView.currentIndex, drag.target.rowSpan,
                                                drag.target.columnSpan, groupId)
                                }
                            }
                        }

                        if (GroupController.groupContainer !== undefined && GroupController.groupContainer !== null) {
                            var itemAtGroupPos = appIcon.mapToItem(GroupController.groupContainer, 0, 0)
                            if (itemAtGroupPos.x % mapContainer.width < -appIcon.width ||
                                    itemAtGroupPos.x > mapContainer.width ||
                                    itemAtGroupPos.y % mapContainer.height < -appIcon.height ||
                                     itemAtGroupPos.y > mapContainer.height) {
                                changeContainerPageTimer.stop()
                                dragMousearea.isChangeContainerPage = false
                                if (GroupController.groupPopuped) {
                                    itemToDesktopTimer.start()
                                }
                            } else if (GroupController.groupPopuped && dragMousearea.Drag.active &&
                                    itemAtGroupPos.x % mapContainer.width > mapContainer.width - appIcon.width + 32 * Style.tabletScaleFactor) {
                                changeContainerPageTimer.start()
                                itemToDesktopTimer.stop()
                                if (isChangeContainerPage && mapContainer.currentIndex < mapContainer.count - 1) {
                                    mapContainer.currentIndex++;
                                    modelManager.removePlaceholderItem()
                                    // 拖动至不同页时才添加 placeholder item
                                    var groupPageItemCount = modelManager.groupPageItemCount(GroupController.groupContainerId, mapContainer.currentIndex);
                                    var itemAtGroupIndex = groupPageItemCount > 0 ? groupPageItemCount : 0
                                    modelManager.addPlaceholderItem(itemAtGroupIndex, mapContainer.currentIndex, drag.target.rowSpan,
                                                drag.target.columnSpan, GroupController.groupContainerId)
                                    isChangeContainerPage = false
                                    changeContainerPageTimer.restart()
                                }
                            } else if (GroupController.groupPopuped && dragMousearea.Drag.active &&
                                       itemAtGroupPos.x % mapContainer.width < -32 * Style.tabletScaleFactor) {
                                changeContainerPageTimer.start()
                                itemToDesktopTimer.stop()
                                if (isChangeContainerPage && mapContainer.currentIndex > 0) {
                                    mapContainer.currentIndex--;
                                    modelManager.removePlaceholderItem()
                                    groupPageItemCount = modelManager.groupPageItemCount(GroupController.groupContainerId, mapContainer.currentIndex);
                                    itemAtGroupIndex = groupPageItemCount > 0 ? groupPageItemCount : 0
                                    modelManager.addPlaceholderItem(itemAtGroupIndex, mapContainer.currentIndex, drag.target.rowSpan,
                                                drag.target.columnSpan, GroupController.groupContainerId)
                                    isChangeContainerPage = false
                                    changeContainerPageTimer.restart()
                                }
                            }
                        }
                    }
                    onReleased: {
                        console.log("Launcher view release.")
                        popUpTimer.stop()
                        mouseAreaReleased = true
                        appIcon.width = type === LauncherItem.Group ? 88 * Style.tabletScaleFactor : 96 * Style.tabletScaleFactor
                        appIcon.height = type === LauncherItem.Group ? 88 * Style.tabletScaleFactor : 96 * Style.tabletScaleFactor
                        isScrolling = false
                        if (dragMousearea.Drag.active) {
                            dragMousearea.Drag.active = false
                            dragMousearea.drag.target = null
                            if (taskbar.replayAllowed) {
                                if (modelData.type === LauncherItem.AppIcon) {
                                    if (modelManager.replacePlaceholderItemTaskbar(visualIndex, pageIndex, -1, 0)) {
                                        appIcon.visible = false
                                        appName.visible = false
                                    }
                                }
                                taskbar.replayAllowed = false
                            } else if (GroupController.aboutToInsertGroup) {
                                GroupController.insertItemToGroup(visualIndex, pageIndex, modelData.id, appIcon, appName)
                            } else {
                                modelManager.replacePlaceholderItem(visualIndex, pageIndex)
                            }
                            modelManager.updateCachedData()
                            modelManager.updateCachedData(0)
                        }
                    }
                }
                Rectangle {
                    id: item
                    property bool moving: listView.horizontalVelocity
                    width: modelData.columnSpan * Style.appIconGridWidth - Style.tabletScaleFactor * 40
                    height: modelData.rowSpan * Style.appIconGridHeight - Style.tabletScaleFactor * 40
                    color: "transparent"
                    clip: false
                    radius: 5
                    anchors {
                        horizontalCenter: parent.horizontalCenter;
                        verticalCenter: parent.verticalCenter
                    }

                    AppIconContainer {
                        id: appIcon
                        anchors.centerIn: parent
                        width: type === LauncherItem.Group ? 88 * Style.tabletScaleFactor : 96 * Style.tabletScaleFactor
                        height: type === LauncherItem.Group ? 88 * Style.tabletScaleFactor : 96 * Style.tabletScaleFactor
                        rootItem: pluginAnchorItem
                        dropShadowOffset: 3
                        deleteIconWidth: 40
                        allowGroupItemToTaskbar: taskbar.replayAllowed
                        widgetColumn: modelData.columnSpan
                        widgetRow: modelData.rowSpan
                        opacity: {
                            if (modelData.type === LauncherItem.AppIcon) {
                                if (listView.moving || dragMousearea.Drag.active) {
                                    return 1
                                }
                                if (dragMousearea.pressed && !dragMousearea.Drag.active) {
                                    return 0.5
                                }
                            }
                            return 1
                        }
                        icon: {
                            switch (modelData.type) {
                            case LauncherItem.AppIcon: return "image://icons/" + modelData.icon
                            case LauncherItem.AddIcon:
                                if (dragMousearea.pressed) {
                                    return "qrc:/ui/images/list-add-click.svg";
                                } else {
                                    return "qrc:/ui/images/list-add-default.svg";
                                }
                            default: return ""
                            }
                        }
                        type: modelData.type
                        containerState: modelData.state
                        groupModel: modelData.model
                        groupName: modelData.name
                        itemId: modelData.id
                        onDeleteIconSig: {
                            appIcon.visible = true

                            UninstallJs.createUnInstallDialog(root, modelData.desktopFile, modelData.name, modelData.icon)

                            taskbar.textPromptBox = model.modelData.name
                            taskbar.iconPromptBox = modelData.icon
                        }
                        onGroupItemToDesktop: {
                            if (!modelManager.itemExistsOnDesktop(itemId, pageIndex) && itemId !== -1) {
                                modelManager.addPlaceholderItem(modelManager.groupPageItemCount(-1, listView.currentIndex), listView.currentIndex)
                            }
                        }
                    }
                    AppNameText {
                        id: appName
                        anchors.top: appIcon.bottom
                        anchors.horizontalCenter: appIcon.horizontalCenter
                        anchors.topMargin: 6 * Style.tabletScaleFactor
                        text: modelData.name
                        width: appNameWidth
                        visible: modelData.type !== LauncherItem.Placeholder &&
                                 modelData.nameVisible &&
                                 appIcon.containerState !== LauncherItem.Dragging &&
                                 modelData.type !== LauncherItem.Widget
                    }
                }

                Component.onCompleted: {
                    root.z = 1
                    if (viewMoel === null) {
                        console.warn("AppContainer: The model is null, please assign a model!")
                    }
                }
            }
        }
    }
    TaskBar {
        id: taskbar
        parentChange: root
        anchors.horizontalCenter: listView.horizontalCenter
        anchors.bottom: listView.bottom
        anchors.bottomMargin: 36 * Style.tabletScaleFactor
        onTaskbarItemEnterDesktop: {
            modelManager.addPlaceholderItem(modelManager.groupPageItemCount(-1, listView.currentIndex), listView.currentIndex)
            taskbarReplayAllowed = true
        }
        onTaskbarItemEnterTaskbar: {
            modelManager.removePlaceholderItem()
            taskbarReplayAllowed = false
        }
        onGroupItemEnterDesktop: {
            modelManager.addPlaceholderItem(modelManager.groupPageItemCount(-1, listView.currentIndex), listView.currentIndex)
        }
    }

    ParallelAnimation {
        running: GroupController.groupPopuped
        PropertyAnimation { target: listView; property: "scale"; from: 1; to: 0.9; easing.type: Easing.OutCubic; duration: 150; }
        PropertyAnimation { target: taskbar; property: "scale"; from: 1; to: 0.9; easing.type: Easing.OutCubic; duration: 150; }
    }
    ParallelAnimation {
        running: !GroupController.groupPopuped
        PropertyAnimation { target: listView; property: "scale"; from: 0.9; to: 1; easing.type: Easing.OutCubic; duration: 150; }
        PropertyAnimation { target: taskbar; property: "scale"; from: 0.9; to: 1; easing.type: Easing.OutCubic; duration: 150; }
    }
}






